const { ipcMain } = require('electron');

// Replace require with dynamic import
async function initializeRealtimeClient(apiKey, getRealtimeWindow) {
  if (!apiKey) {
    console.error('OpenAI API key is required to initialize RealtimeClient');
    return;
  }

  const { RealtimeClient } = await import('@openai/realtime-api-beta');

  let realtimeWindow = getRealtimeWindow();
  console.log("Initializing with API key.");

  // Initialize the RealtimeClient
  const client = new RealtimeClient({
    apiKey: apiKey,
  });

  // Configure the client with your desired settings
  client.updateSession({
    instructions: 'You are a helpful assistant.',
    voice: 'alloy',
    turn_detection: { type: 'server_vad' }, // Enable server-side VAD
    input_audio_transcription: { model: 'whisper-1' },
  });

  // Connect the client
  await client.connect();

  // Handle IPC events from the renderer process

  // Handle audio data sent from the renderer
  ipcMain.on('realtime-send-audio', (event, audioData) => {
    if (client) {
      // Convert the received data to Int16Array if necessary
      let int16Array;
      if (audioData instanceof Int16Array) {
        int16Array = audioData;
      } else if (audioData.buffer) {
        int16Array = new Int16Array(audioData.buffer);
      } else {
        console.error('Received audio data is not in Int16Array format.');
        return;
      }
      client.appendInputAudio(int16Array);
    }
  });

  // Handle user text input (if applicable)
  ipcMain.on('realtime-send-user-input', (event, text) => {
    if (client) {
      client.sendUserMessageContent([{ type: 'input_text', text }]);
    }
  });

  // Handle manual response creation (if needed)
  ipcMain.on('realtime-create-response', () => {
    console.log("createResponse");
    if (client) {
      client.createResponse();
    }
  });

  // Handle cancel event to interrupt the AI
  ipcMain.on('realtime-send-cancel', () => {
    if (client) {
      client.cancelResponse();
    }
  });

  // Handle conversation updates and send them to the renderer process
  client.on('conversation.updated', ({ item, delta }) => {
    console.log("Conversation updated:", item);
    if (realtimeWindow && !realtimeWindow.isDestroyed()) {
      realtimeWindow.webContents.send('realtime-conversation-updated', { item, delta });
    }
  });

  // Handle errors and send them to the renderer process
  client.on('error', (error) => {
    console.error('Realtime API Error:', error);
    if (realtimeWindow && !realtimeWindow.isDestroyed()) {
      realtimeWindow.webContents.send('realtime-error', error);
    }
  });

  // Optionally, handle disconnection
  realtimeWindow.on('closed', () => {
    if (client) {
      client.disconnect();
    }
  });
}

module.exports = initializeRealtimeClient;
